<?php

namespace Drupal\datalist\Element;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element;
use Drupal\Core\Render\Element\FormElement;
use Drupal\Core\Render\Element\Textfield;

/**
 * Provides a render element for a datalist input.
 *
 * It will render a textfield connected to a datalist.
 * Currently multivalue and grouping is not supported.
 * By default we hide the keys when rendering.
 *
 * Usage example:
 * @code
 * $form['author']['name'] = array(
 *   '#type' => 'datalist',
 *   '#options' => [],
 *   '#title' => '',
 *   '#required' => TRUE|FALSE,
 *   '#placeholder' => 'a placeholder for the input field',
 *   '#default_value' => ''
 *   '#clear_button' => FALSE|"string for the clear button label" (defaults to 'X')
 *   '#clear_button_description' => FALSE|"string for describing the clear button" (defaults to 'Clear field')
 *   '#use_keys' => TRUE|FALSE (default=FALSE)
 *   '#allow_fallback' => TRUE|FALSE ==> when browser doesnt support datalist fallback to autocomplete, Be sure to fill in autocomplete.
 *   '#autocomplete_route_name' => NULL, OPTIONAL
 *   '#autocomplete_route_parameters' => [], OPTIONAL
 * );
 * @endcode
 *
 * @FormElement("datalist")
 */
class Datalist extends Textfield {

  /**
   * @param $element
   * @param bool $isSupportedBrowser
   * @return array
   */
  public static function fallbackTextfield($element, bool $isSupportedBrowser): array {
    // Reset to base information.
    $route = $element['#autocomplete_route_name'];
    $element = array_merge($element, $element['#parent_textfield']);
    $element['#autocomplete_route_name'] = $route;

    $element['#type'] = 'textfield';

    $element['#supported_browser'] = $isSupportedBrowser;

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = static::class;
    return [
        '#input' => TRUE,
        '#process' => [
          [$class, 'processDatalist'],
          [$class, 'processAutocomplete'],
          [$class, 'processAjaxForm'],
          [$class, 'processPattern'],
          [$class, 'processGroup'],
        ],
        '#pre_render' => [
          [$class, 'preRenderTextfield'],
          [$class, 'preRenderDatalist'],
          [$class, 'preRenderGroup'],
        ],
        '#theme' => 'input__datalist',
        '#options' => [],
        '#clear_button' => '✖',
        '#down_button' => '▼',
        '#clear_button_description' => t('Clear field'),
        '#use_keys' => FALSE,
        '#autocomplete_route_name' => NULL,
        '#autocomplete_route_parameters' => [],
        '#parent_textfield' => parent::getInfo(),
        '#allow_fallback' => TRUE,
      ] + parent::getInfo();
  }

  /**
   * @inheritdoc
   */
  public static function processDatalist(&$element, FormStateInterface $form_state, &$complete_form) {
    $isSupportedBrowser = Datalist::isSupportedBrowser($element);
    if (!$isSupportedBrowser) {
      return self::fallbackTextfield($element, $isSupportedBrowser);
    }

    $element['#list']  = "{$element['#id']}-datalist";

    if (!empty($element['#clear_button'])) {
      $element['#attached']['library'][] = 'datalist/datalist';
    }

    // Let validation on empty value also work by adding the empty value.
    // Since we add #options, core will perform a required validation on our
    // submitted value. Since datalist is basically an enhanced  textfield, when
    // the form is submitted without interacting with the field, the validation
    // method will receive an empty string. In order for this string to be
    // considered legal, we always add it to our options list. When a field is
    // required, it will now get a required error and not an illegal choice.
    // In preprocess we unset this empty choice if the field is required.
    $element['#options'] =  ['' => $element['#placeholder'] ?? ''] + $element['#options'];

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public static function valueCallback(&$element, $input, FormStateInterface $form_state) {
    if ($input !== FALSE && $input !== NULL) {
      // We might just have submitted without selecting or typing anything.
      // In that case an empty string is used as input.
      if ($input === '' || $element['#use_keys']) {
        return $input;
      }
      else {
        // If the input is coming from the textinput, and we don't #use_keys as
        // values, we are getting the label as a value, so lookup the key for
        // this label.
        $lookup = array_search($input, $element['#options'], true);

        if ($lookup !== FALSE) {
          return $lookup;
        }

        // Setting the input in code will result in the input being the key
        // already, so in case we don't find it just return what was given.
        if (key_exists($input, $element['#options'])) {
          return $input;
        }

      }
    }

    return NULL;
  }

  /**
   * Prepares a datalist render element.
   */
  public static function preRenderDatalist($element) {

    if (isset($element['#supported_browser']) && !$element['#supported_browser']) {
      return $element;
    }

    // Make sure we keep the datalist type class, even if we extend this class.
    $element['#wrapper_attributes']['class'][] = 'form-type--datalist';

    if (!($element['#required'] ?? FALSE)) {
      // Don't show the empty value, it's shown by the placeholder.
      unset($element['#options']['']);
    }

    if (!$element['#use_keys'] && $element['#value'] !== '') {
      // Transform the actual field value which is a key into a label for displaying.
      $element['#value'] = $element['#options'][$element['#value']] ?? '';
    }

    $element['#attributes']['class'][] = 'form-element';
    $element['#attributes']['type'] = 'text';
    $element['#attributes']['autocomplete'] = 'off';
    Element::setAttributes($element, ['id', 'name', 'value', 'size', 'maxlength', 'placeholder', 'list', 'autocomplete']);
    static::setAttributes($element, ['form-text', 'form-datalist']);

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public static function preRenderAjaxForm($element) {
    if (!empty($element['#ajax']) && !isset($element['#ajax_processed']) && !isset($element['#ajax']['event'])) {
      $element['#ajax']['event'] = 'change';
    }
    return parent::preRenderAjaxForm($element);
  }

  public static function isSupportedBrowser($element) {
    $userAgent = $_SERVER['HTTP_USER_AGENT'];
    
    $isFirefoxMobileAndroid = (str_contains($userAgent, 'Mozilla/5.0 (Android')
      && str_contains($userAgent, 'Firefox/')
      && str_contains($userAgent, 'Gecko/'));
    $isFirefoxMobileIOS = (str_contains($userAgent, 'Mozilla/5.0 (iP')
      && str_contains($userAgent, 'FxiOS/1')
      && str_contains($userAgent, 'AppleWebKit/'));

    $isEdgeMobileAndroid = (str_contains($userAgent, 'Mozilla/5.0 (Android')
      && str_contains($userAgent, 'AppleWebKit/')
      && str_contains($userAgent, 'Edge/'));
    $isEdgeMobileIOS = (str_contains($userAgent, 'Mozilla/5.0 (Android')
      && str_contains($userAgent, 'AppleWebKit/')
      && str_contains($userAgent, 'EdgiOS/'));

    return !$isEdgeMobileAndroid
      && !$isEdgeMobileIOS
      && !$isFirefoxMobileAndroid
      && !$isFirefoxMobileIOS
      && $element['#allow_fallback'];
  }

  /**
   * Adds autocomplete functionality to elements.
   *
   * This sets up autocomplete functionality for elements with an
   * #autocomplete_route_name property, using the #autocomplete_route_parameters
   * and #autocomplete_query_parameters properties if present.
   *
   * For example, suppose your autocomplete route name is
   * 'mymodule.autocomplete' and its path is
   * '/mymodule/autocomplete/{a}/{b}'. In a form array, you would create a text
   * field with properties:
   * @code
   * '#autocomplete_route_name' => 'mymodule.autocomplete',
   * '#autocomplete_route_parameters' => array('a' => $some_key, 'b' => $some_id),
   * @endcode
   * If the user types "keywords" in that field, the full path called would be:
   * 'mymodule_autocomplete/$some_key/$some_id?q=keywords'
   *
   * @param array $element
   *   The form element to process. Properties used:
   *   - #autocomplete_route_name: A route to be used as callback URL by the
   *     autocomplete JavaScript library.
   *   - #autocomplete_route_parameters: The parameters to be used in
   *     conjunction with the route name.
   *   - #autocomplete_query_parameters: The parameters to be used in
   *     query string
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param array $complete_form
   *   The complete form structure.
   *
   * @return array
   *   The form element.
   */
  public static function processAutocomplete(&$element, FormStateInterface $form_state, &$complete_form) {
    if (isset($element['#supported_browser']) && !$element['#supported_browser']) {
      return Textfield::processAutocomplete($element, $form_state, $complete_form);
    }
    return $element;
  }

  /**
   * #process callback for #pattern form element property.
   *
   * @param array $element
   *   An associative array containing the properties and children of the
   *   generic input element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param array $complete_form
   *   The complete form structure.
   *
   * @return array
   *   The processed element.
   */
  public static function processPattern(&$element, FormStateInterface $form_state, &$complete_form) {
    if (isset($element['#supported_browser']) && !$element['#supported_browser']) {
      return Textfield::processPattern($element, $form_state,$complete_form);
    }
    return $element;
  }

  /**
   * Prepares a #type 'textfield' render element for input.html.twig.
   *
   * @param array $element
   *   An associative array containing the properties of the element.
   *   Properties used: #title, #value, #description, #size, #maxlength,
   *   #placeholder, #required, #attributes.
   *
   * @return array
   *   The $element with prepared variables ready for input.html.twig.
   */
  public static function preRenderTextfield($element) {
    if (isset($element['#supported_browser']) && !$element['#supported_browser']) {
      return Textfield::preRenderTextfield($element);
    }
    return $element;
  }

}
